package flinktemplistener;

import flinktemplistener.events.TemperatureEvent;
import flinktemplistener.utility.AggregateStreamPublisher;
import flinktemplistener.utility.AverageTempPublisher;
import org.apache.flink.api.common.functions.MapFunction;
import org.apache.flink.api.java.tuple.Tuple2;
import org.apache.flink.api.java.tuple.Tuple3;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.windowing.assigners.GlobalWindows;
import org.apache.flink.streaming.api.windowing.triggers.CountTrigger;

import org.slf4j.LoggerFactory;
import org.slf4j.Logger;
        
/**
 * @author ebruno
 */
public class FlinkAverageTemperature {
    private String pubSubServerAddr  = "localhost";
    private Integer pubSubServerPort = 4222;
    private int deviceId = 1;
    private int devicePort = 9091;
    static Logger logger = LoggerFactory.getLogger(FlinkAverageTemperature.class);;
    
    public static void main(String[] args) throws Exception {
        String host = "localhost";
        int port = 4222;
        int deviceId = 1;
        int devicePort = 9091;
        
        if ( args != null ) {
            String param = "";
            for ( String arg: args ) {
                switch ( param ) {
                    case "host":
                        host = arg; // PubSub host
                        param = "";
                        break;
                        
                    case "port":
                        port = new Integer(arg); // PubSub port
                        param = "";
                        break;
                        
                    case "devicePort":
                        devicePort = new Integer(arg);
                        param = "";
                        break;
                        
                    case "deviceid":
                        deviceId = new Integer(arg);
                        param = "";
                        break;
                }
                
                // Save the param name to process the value next time around
                param = arg;
            }
        }

        FlinkAverageTemperature listener = 
                new FlinkAverageTemperature(host, port, deviceId, devicePort);

        Thread tempAverageThread = listener.startAverageTemperatureReporting();
        
        tempAverageThread.join();
    }
    
    public FlinkAverageTemperature(String host, int port, int deviceId, int devicePort) {
        this.pubSubServerAddr = host;
        this.pubSubServerPort = port;
        this.deviceId = deviceId;
        this.devicePort = devicePort;
    }
    
    private Thread startAverageTemperatureReporting() {
        Thread t = new Thread() {
            public void run() {
                try {
                    monitorAverageTemperature();
                }
                catch ( Exception e ) {
                    e.printStackTrace();
                }
            }
        };

        t.start();

        return t;
    }
    
    private void monitorAverageTemperature() throws Exception {
        StreamExecutionEnvironment env = 
                StreamExecutionEnvironment.getExecutionEnvironment();

        //
        // 1-put the temp events into the stream
        // 2-select all from the stream and calc temp count and sum
        // 3-select all from the stream and continuously average the temps
        // 4-publish the values
        //

        // Connect to device and create a stream of temperature data
        //
        int key = this.deviceId;
        DataStream<String> deviceRawStream = env.socketTextStream("localhost", devicePort);
        DataStream<TemperatureEvent> inputEventStream = 
            deviceRawStream.map(
                new MapFunction<String, TemperatureEvent>() {
                    @Override
                    public TemperatureEvent map(String value) throws Exception {
                        TemperatureEvent evt = 
                                new TemperatureEvent(key, new Integer(value));
                        return evt;
                    }
                }
            );

        //
        // Each of the following "average" processes runs at different time
        // intervals but with the same set of data, so the answers all equal
        // even though they run and output the results at different times
        //

        // Aggregates all temperature data to facilitate a federated
        // global average temperature across all devices
        //
        SingleOutputStreamOperator<Tuple3<Integer, Long, Double>> aggregateProcess =
                    inputEventStream.keyBy( value -> value.getDeviceId() )
                    .window(GlobalWindows.create())
                    .trigger(CountTrigger.of(1))
                    .aggregate(new Aggregation());

        // Publish this stream over NATS.io Pub/Sub for distribution/federation
        //
        aggregateProcess.addSink(
                new AggregateStreamPublisher(
                        pubSubServerAddr, 
                        pubSubServerPort, 
                        "AggregateTemp"+deviceId) );
        aggregateProcess.print();
        
        // Processes the average temp with each event that arrives
        //
        SingleOutputStreamOperator<Tuple2<Integer, Double>> countAvgProcess =
                    inputEventStream.keyBy( value -> value.getDeviceId() )
                    .window(GlobalWindows.create())
                    .trigger(CountTrigger.of(1))
                    .aggregate(new Average());

        // Publish this stream over NATS.io Pub/Sub for distribution/federation
        //
        countAvgProcess.addSink(
                new AverageTempPublisher(
                        pubSubServerAddr, 
                        pubSubServerPort, 
                        "AverageTemp"+deviceId) );
        countAvgProcess.print();

        /*
        // Processes the average temp every 2 seconds
        //
        SingleOutputStreamOperator<Tuple2<Integer, Double>> timeAvgProcess =
                    inputEventStream.keyBy( value -> value.getDeviceId() )
                    .window(GlobalWindows.create())
                    .trigger(ContinuousProcessingTimeTrigger.of(Time.milliseconds(2000)))
                    .aggregate(new Average());
        timeAvgProcess.print();
        */
        
        System.out.println("Calling env.execute() for Temperature Averaging");
        env.execute("Device Temperature");
    }
}

