package com.ebruno.natsiopubsubconnector;

import io.nats.client.Connection;
import io.nats.client.ExceptionHandler;
import io.nats.client.Message;
import io.nats.client.NATSException;
import io.nats.client.Nats;
import io.nats.client.Options;
import io.nats.client.SyncSubscription;

import org.apache.flink.api.common.serialization.DeserializationSchema;
import org.apache.flink.api.common.io.ratelimiting.FlinkConnectorRateLimiter;
import org.apache.flink.api.common.typeinfo.TypeInformation;
import org.apache.flink.api.java.typeutils.ResultTypeQueryable;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.runtime.state.CheckpointListener;
import org.apache.flink.streaming.api.checkpoint.ListCheckpointed;
import org.apache.flink.streaming.api.functions.source.ParallelSourceFunction;
import org.apache.flink.streaming.api.functions.source.RichSourceFunction;
import org.apache.flink.util.Preconditions;

import java.util.List;
import java.io.IOException;
import java.util.HashMap;
//import scala.collection.mutable.Subscriber;

/**
 * @author ebruno
 */
public class PubSubSource<OUT> extends RichSourceFunction<OUT> 
    implements  ResultTypeQueryable<OUT>, 
                ParallelSourceFunction<OUT>, 
                CheckpointListener, 
                ListCheckpointed<AcknowledgeIdsForCheckpoint<String>>,
                ExceptionHandler {

	protected final PubSubDeserializationSchema<OUT> deserializationSchema;
	protected final FlinkConnectorRateLimiter rateLimiter;
	protected final int messagePerSecondRateLimit;
	protected Boolean isRunning = true;
    protected transient Connection natsConn;
    protected transient SyncSubscription sub;
    protected Boolean connected = false;
    
    protected String server = "localhost";
    protected Integer port = 4222;
    protected String topic;
    protected String username;
    protected String password;
    protected static HashMap<String, SyncSubscription> subs = new HashMap<>();

	PubSubSource(PubSubDeserializationSchema<OUT> deserializationSchema,
                String server,
                int port,
                String username,
                String password,
                String subscriptionName,
				FlinkConnectorRateLimiter rateLimiter,
				int messagePerSecondRateLimit) {
		this.deserializationSchema = deserializationSchema;
		this.rateLimiter = rateLimiter;
		this.messagePerSecondRateLimit = messagePerSecondRateLimit;
        this.topic = subscriptionName;
        this.server = server;
        this.port = port;
        this.username = username;
        this.password = password;
	}

    private void connectToNATSio() {
        synchronized ( subs ) {
            try {
                if ( subs.containsKey(topic) ) {
                    sub = subs.get(topic);
                    return;
                }

                Options.Builder builder = 
                        new Options.Builder()
                                .pingInterval(10000)
                                .timeout(10000)
                                .reconnectWait(1000)
                                .errorCb(this)
                                .verbose();

                String hostname =  "nats://"+server+":"+port;

                natsConn = Nats.connect( hostname, builder.build() );
                sub = natsConn.subscribe(topic);
                subs.put(topic, sub);

                System.out.println("Subscribed to "+topic+" on host " + hostname);

                isRunning = true;
                connected = true;

                if ( ! sub.isValid() ) {
                    isRunning = false;
                }
            }
            catch ( Exception e ) {
                e.printStackTrace();
            }
        }
    }
    
	@Override
	public void open(Configuration configuration) throws Exception {
		super.open(configuration);
        
        getRuntimeContext().getMetricGroup().gauge(
                "PubSubMessagesProcessedNotAcked", 
                this::getOutstandingMessagesToAck);

		//convert per-subtask-limit to global rate limit, as FlinkConnectorRateLimiter::setRate expects a global rate limit.
		if ( rateLimiter != null ) {
            rateLimiter.setRate( messagePerSecondRateLimit * getRuntimeContext().getNumberOfParallelSubtasks() );
            rateLimiter.open( getRuntimeContext() );
        }

        if ( ! connected ) {
            connectToNATSio();
        }
    }

	@Override
    public void close() throws Exception {
        System.out.println("#### Un-subscribed to "+topic);
        if ( sub != null ) {
            sub.close();
        }

        if ( natsConn != null ) {
            natsConn.close();
        }
    }
    
	@Override
	public void run(SourceContext<OUT> sourceContext) throws Exception {
		while (isRunning) {
			try {
                Message msg = sub.nextMessage(1 * 60 * 1000 * 60);
                synchronized ( sourceContext.getCheckpointLock() ) {
//    				acknowledgeOnCheckpoint.addAcknowledgeId(message.getAckId());
                    OUT deserializedMessage = deserializationSchema.deserialize(msg);
                    if (deserializationSchema.isEndOfStream(deserializedMessage)) {
                        cancel();
                        return;
                    }

                    sourceContext.collect(deserializedMessage);
                }
			} 
            catch ( Exception e ) {
                e.printStackTrace();
				//isRunning = false;
			}
		}
        
		sub.close();
	}

	private Integer getOutstandingMessagesToAck() {
        return 0;
		//return acknowledgeOnCheckpoint.numberOfOutstandingAcknowledgements();
	}

    @Override
    public void cancel() {
        isRunning = false;
    }

    @Override
    public TypeInformation<OUT> getProducedType() {
        return null;
    }

    @Override
    public void notifyCheckpointComplete(long checkpointId) throws Exception {
    }

    @Override
    public List<AcknowledgeIdsForCheckpoint<String>> snapshotState(long checkpointId, long timestamp) throws Exception {
        return null;
    }

    @Override
    public void restoreState(List<AcknowledgeIdsForCheckpoint<String>> state) throws Exception {
    }

    @Override
    public void onException(NATSException natse) {
        System.out.println( natse.toString() );
    }
    
	public static DeserializationSchemaBuilder newBuilder() {
		return new DeserializationSchemaBuilder();
	}
    
	/**
	 * Builder to create PubSubSource.
	 *
	 * @param <OUT> The type of objects which will be read
	 */
	public static class PubSubSourceBuilder<OUT> implements ProjectNameBuilder<OUT>, SubscriptionNameBuilder<OUT> {
		private PubSubDeserializationSchema<OUT> deserializationSchema;
		private String server = "localhost";
        private int port = 4222;
        private String projectName;
		private String subscriptionName;
        private String username, password; // credetials
                
		private PubSubSubscriberFactory pubSubSubscriberFactory;
		//private Credentials credentials;
		private int maxMessageToAcknowledge = 10000;
		private int messagePerSecondRateLimit = 100000;

		private PubSubSourceBuilder(
                DeserializationSchema<OUT> deserializationSchema) {
			Preconditions.checkNotNull(deserializationSchema);
			this.deserializationSchema = new PubSubDeserializationSchemaImpl<>(deserializationSchema);
		}

		private PubSubSourceBuilder(
                PubSubDeserializationSchema<OUT> deserializationSchema) {
			Preconditions.checkNotNull(deserializationSchema);
			this.deserializationSchema = deserializationSchema;
		}

		@Override
		public SubscriptionNameBuilder<OUT> withProjectName(String projectName) {
			Preconditions.checkNotNull(projectName);
			this.projectName = projectName;
			return this;
		}

		@Override
		public PubSubSourceBuilder<OUT> withSubscriptionName(String subscriptionName) {
			Preconditions.checkNotNull(subscriptionName);
			this.subscriptionName = subscriptionName;
			return this;
		}

		/**
		 * Set the credentials.
		 * If this is not used then the credentials are picked up from the environment variables.
		 *
		 * @param credentials the Credentials needed to connect.
		 * @return The current PubSubSourceBuilder instance
		 */
		public PubSubSourceBuilder<OUT> withCredentials(String username, String password) {
			this.username = username;
            this.password = password;
			return this;
		}
        
		public PubSubSourceBuilder<OUT> withServer(String server) {
			this.server = server;
			return this;
		}

        public PubSubSourceBuilder<OUT> withport(int port) {
			this.port = port;
			return this;
		}

		/**
		 * Set a PubSubSubscriberFactory
		 * This allows for custom Subscriber options to be set.
		 * {@link DefaultPubSubSubscriberFactory} is the default.
		 *
		 * @param pubSubSubscriberFactory A factory to create a {@link Subscriber}
		 * @return The current PubSubSourceBuilder instance
		 */
		public PubSubSourceBuilder<OUT> withPubSubSubscriberFactory(PubSubSubscriberFactory pubSubSubscriberFactory) {
			this.pubSubSubscriberFactory = pubSubSubscriberFactory;
			return this;
		}

		/**
		 * There is a default PubSubSubscriber factory that uses gRPC to pull in PubSub messages. This method can be used to tune this default factory.
         * Note this will not work in combination with a custom PubSubSubscriber factory.
		 * @param maxMessagesPerPull the number of messages pulled per request. Default: 100
		 * @param perRequestTimeout the timeout per request. Default: 15 seconds
		 * @param retries the number of retries when requests fail
		 * @return The current PubSubSourceBuilder instance
		 */
//		public PubSubSourceBuilder<OUT> withPubSubSubscriberFactory(int maxMessagesPerPull, Duration perRequestTimeout, int retries) {
//			this.pubSubSubscriberFactory = new PubSubSubscriberFactory(ProjectSubscriptionName.format(projectName, subscriptionName),
//																			retries,
//																			perRequestTimeout,
//																			maxMessagesPerPull);
//			return this;
//		}

		/**
		 * Set a limit on the rate of messages per second received. This limit is per parallel instance of the source function.
		 * Default is set to 100000 messages per second.
		 *
		 * @param messagePerSecondRateLimit the message per second rate limit.
		 */
		public PubSubSourceBuilder<OUT> withMessageRateLimit(int messagePerSecondRateLimit) {
			this.messagePerSecondRateLimit = messagePerSecondRateLimit;
			return this;
		}

		/**
		 * Actually build the desired instance of the PubSubSourceBuilder.
		 *
		 * @return a brand new SourceFunction
		 * @throws IOException              in case of a problem getting the credentials
		 * @throws IllegalArgumentException in case required fields were not specified.
		 */
		public PubSubSource<OUT> build() throws IOException {
            return new PubSubSource(
                    deserializationSchema,
                    server, port,
                    username, password,
                    subscriptionName, // projectName+subscriptionName
                    null,
                    10000);
		}
	}

    /**
	 * Part of {@link PubSubSourceBuilder} to set required fields.
	 */
	public static class DeserializationSchemaBuilder {
		/**
		 * Set the DeserializationSchema used to deserialize incoming PubSubMessages.
		 * If you want access to meta data of a PubSubMessage use the overloaded withDeserializationSchema({@link PubSubDeserializationSchema}) method instead.
		 */
		public <OUT> ProjectNameBuilder<OUT> withDeserializationSchema(DeserializationSchema<OUT> deserializationSchema) {
			return new PubSubSourceBuilder<>(deserializationSchema);
		}

		/**
		 * Set the DeserializationSchema used to deserialize incoming PubSubMessages.
		 */
		public <OUT> ProjectNameBuilder<OUT> withDeserializationSchema(PubSubDeserializationSchema<OUT> deserializationSchema) {
			return new PubSubSourceBuilder<>(deserializationSchema);
		}
	}

	/**
	 * Part of {@link PubSubSourceBuilder} to set required fields.
	 */
	public interface ProjectNameBuilder<OUT> {
		/**
		 * Set the project name of the subscription to pull messages from.
		 */
		SubscriptionNameBuilder<OUT> withProjectName(String projectName);
	}

	/**
	 * Part of {@link PubSubSourceBuilder} to set required fields.
	 */
	public interface SubscriptionNameBuilder<OUT> {
		/**
		 * Set the subscription name of the subscription to pull messages from.
		 */
		PubSubSourceBuilder<OUT> withSubscriptionName(String subscriptionName);
	}

//	static class AcknowledgeOnCheckpointFactory implements Serializable {
//		AcknowledgeOnCheckpoint<String> create(Acknowledger<String> acknowledger) {
//			return new AcknowledgeOnCheckpoint<>(acknowledger);
//		}
//	}
    
    
}
